#!/usr/bin/python
# vim: ts=4:et
"""Watches the filesystem modifications."""

import sys
import httpserver
import time

from utils.logger import log

import pyinotify

import yaml
import syslog
import threading
import errno

import os
import posixpath
import posix
import stat

from repository import Repository
from webserver import GsyncWebserver
import httpserver

GSYNC_MASK = (
    pyinotify.IN_MODIFY |
    pyinotify.IN_CLOSE_WRITE |
    pyinotify.IN_ATTRIB |
    pyinotify.IN_MOVED_FROM |
    pyinotify.IN_MOVED_TO |
    pyinotify.IN_CREATE |
    pyinotify.IN_DELETE |
    pyinotify.IN_Q_OVERFLOW
)

class QueueFull(Exception):
    pass

class OrderedQueue:
    """FIFO queue that groups duplicates"""
    def __init__(self, max=200):
        self.max = max
        self.elems = []
        self.lock = threading.Lock()
        self.sem = threading.Semaphore(0)

    def full(self):
        return len(self.elems) >= self.max
    
    def empty(self):
        return len(self.elems) == 0

    def set_empty(self):
        self.lock.acquire()
        self.elems = []
        self.lock.release()

    def put(self, job):
        if self.full():
            raise QueueFull()

        self.lock.acquire()
        if job not in self.elems:
            self.elems.append(job)
            self.sem.release()
        else:
            pass
        self.lock.release()

    def get(self, blocking=1):
        if self.sem.acquire(blocking):
            self.lock.acquire()
            x = self.elems.pop(0)
            self.lock.release()
            return x

    def find(self, path, ev_mask, cookie=None):
        self.lock.acquire()
        found = None
        for j in self.elems:
            if j.mask & ev_mask:
                if cookie and getattr(j, 'cookie', None) == cookie:
                    found = j
                if not cookie and j.get_path() == path:
                    found = j
        self.lock.release()
        return found

    def changed_later(self, ev):
        self.lock.acquire()
        now = ev
        before = None
        after = None
        cookie = None
        for j in self.elems:
            if j.mask & pyinotify.IN_MOVED_FROM:
                if now and now.get_path().startswith(j.get_path()):
                    cookie = j.cookie
                    if not before:
                        before = now
            if j.mask & pyinotify.IN_MOVED_TO:
                if j.cookie == cookie:
                    now = j
                    after = now
            if j.mask & pyinotify.IN_DELETE:
                if now and now.get_path().startswith(j.get_path()):
                    after = j
                    now = None
        self.lock.release()
        return (before, after)

class GsyncProtoClient:
    """This observes the local filesystem changes taking
    messages from the ProcessEvent class below, and bridges
    this to Repositories using the gsync protocol"""

    def __init__(self, repo):
        self.repo = repo

    def on_change(self, path, src=None):
        log.debug('on_change %s %s (%s)' % (path, self.repo, src,))
        st = posix.lstat(src or path)
        metas = {
            'mode' : st.st_mode,
            'uid': st.st_uid,
            'gid' : st.st_gid,
            'atime' : st.st_atime,
            'mtime' : st.st_mtime,
            '-p': True
        }
        if stat.S_ISREG(st.st_mode):
            fd = file(src or path)
            self.repo.add_from_fd(path, fd, metas)
        elif stat.S_ISDIR(st.st_mode):
            self.repo.update_metas(path, metas)
        elif stat.S_ISLNK(st.st_mode):
            self.repo.update_metas(path, metas)

    def on_delete(self, path):
        log.debug('on_del %s (%s)' % (path, self.repo,))
        self.repo.remove(path)

    def on_create(self, path, src=None):
        log.debug('on_create %s %s (%s)' % (path, self.repo, src,))
        st = posix.lstat(src or path)
        metas = {
            'mode' : st.st_mode,
            'uid': st.st_uid,
            'gid' : st.st_gid,
            'atime' : st.st_atime,
            'mtime' : st.st_mtime,
            '-p': True
        }

        if stat.S_ISREG(st.st_mode):
            f = file(src or path, 'r')
            self.repo.add_from_fd(path, f, metas)
        elif stat.S_ISDIR(st.st_mode):
            self.repo.mkdir(path, metas)
        elif stat.S_ISLNK(st.st_mode):
            self.repo.symlink(path, posix.readlink(path), metas)

    def on_attrib(self, path, src=None):
        log.debug('on_attrib %s %s (%s)' % (path, self.repo, src,))
        st = posix.lstat(src or path)
        metas = {
            'mode' : st.st_mode,
            'uid': st.st_uid,
            'gid' : st.st_gid,
            'atime' : st.st_atime,
            'mtime' : st.st_mtime,
        }
        self.repo.update_metas(path, metas)

    def on_rename(self, a, b):
        self.repo.rename(a, b)

    def sync(self, path):
        """Just check if this bridges has the path"""
        if not self.repo.has(path):
            log.debug('sync: repo has not %s' % path)
            return False

        try:
            st = self.repo.stat(path)
            l_st = posix.stat(path)
        except EnvironmentError, e:
            log.debug('sync: stat failed for %s' % path)
            return False
        
        if st.st_mtime == l_st.st_mtime:
            log.debug('sync: matching mtimes for %s' % path)
            return True

        log.debug('sync: %s != %s for %s' % 
            (st.st_mtime, l_st.st_mtime, path)
        )
        return False

    def resync(self, src=None):
        if not src:
            for path in self.repo.conf.get('monitor',[]):
                self.resync(path)
            return

        log.info('resync of %s for %s' % (src, self.repo,))
        src = posixpath.normpath(src)

        if not self.sync(src) and self.repo.wants(src):
            ## If this is a file, juste create it
            self.on_create(src)

        for (dirname, dirs, files) in os.walk(src):
            if not self.repo.has(dirname) and self.repo.wants(dirname):
                self.on_create(dirname)
            for f in dirs + files:
                p = '%s/%s' % (dirname, f,)
                if not self.sync(p) and self.repo.wants(p):
                    if self.repo.has(p):
                        self.on_change(p)
                    else:
                        self.on_create(p)

        log.info('resync of %s done for %s' % (src, self.repo,))

    def wants(self, path):
        return self.repo.wants(path)

    def has(self, path):
        return self.repo.has(path)

    def alive(self):
        try:
            return self.repo.alive()
        except:
            return False

    def snapshot(self):
        return self.repo.snapshot()

    def __repr__(self):
        return '[%s]' % (self.repo,)

    def setup(self):
        if not self.repo.versions():
            log.info('Initial version for %s' % self.repo)
            try:
                self.repo.create()
            except Exception, e:
                log.err('Could not create %s: %s' % (self.repo, e))
                return None
            self.resync()
        else:
            ver = max(self.repo.versions())
            self.repo.use(ver)
            age = time.time() - self.repo.stat('/').st_mtime
            if self.repo.min_interval and age > self.repo.min_interval:
                ## If we commit recent modifications to this old
                ## revision, we'd lose a good backup image. Rather
                ## snapshot it now.
                log.info('Snapshoting old version of %s' % self.repo)
                self.repo.snapshot()
 
class ResyncNeeded(Exception):
    pass

class GsyncWorker(threading.Thread):
    def __init__(self, delay=30):
        threading.Thread.__init__(self)
        self.running = True
        self.rename_cookies = {}
        self.modified_files = {}
        self.queue = OrderedQueue()
        self.delay = delay

    def set_repository(self, repo):
        self.bridge = GsyncProtoClient(repo)

    def snapshot(self):
        log.debug('snapshot for %s: start' % self.bridge)
        self.bridge.snapshot()
        log.debug('snapshot for %s: done' % self.bridge)

    def stop(self):
        self.running = False

    def wants(self, ev):
        return self.bridge.wants(ev.get_path())

    def sleep(self, cnt=60):
        while cnt and self.running:
            cnt -= 1
            time.sleep(1)

    def resync(self, path):
        log.debug('maybe resync %s' % path)
        while (
            path != '/' and 
            self.bridge.wants(posixpath.dirname(path)) and
            not self.bridge.sync(posixpath.dirname(path))
        ):
            path = posixpath.dirname(path)
            log.debug('maybe resync -> %s' % path)
        log.debug('finally resync %s' % path)
        self.bridge.resync(path)
 
    def _run(self):
        self.bridge.setup()
        while self.running:
            while not self.bridge.alive():
                self.sleep(self.delay)
                if not self.running:
                    return

            if self.queue.full():
                log.warn('Queue full, resyncing')
                self.queue.set_empty()
                self.bridge.resync()
                continue

            job = self.queue.get(False)

            need_resync = []
            def flag_resync(p):
                if p not in need_resync:
                    need_resync.append(p)

            while job:
                log.debug('%s processing %s' % (self, job,))
                if job:
                    try:
                        self.process(job)
                    except EnvironmentError, e:
                        if e.errno == errno.ENOENT:
                            flag_resync(job.get_path())
                    except ResyncNeeded, e:
                        flag_resync(job.get_path())
                    except Exception, e:
                        log.err('%s:%s' % (e.__class__, e))
                job = self.queue.get(False)

            try:
                for path in need_resync:
                    log.debug('need_resync %s' % path)
                    self.resync(path)
            except Exception, e:
                log.err('%s:%s' % (e.__class__, e))

            log.debug('sleeping')
            self.sleep(self.delay)

    def process(self, ev):
        path = ev.get_path()

        ## Try to avoid a race with renames and unlinks
        (before, after) = self.queue.changed_later(ev)
        source = None
        if before:
            if after.mask & pyinotify.IN_DELETE:
                ## Just ignore this case
                return
            elif after.mask & pyinotify.IN_MOVED_TO:
                ## Do everything as usual but reading the
                ## file from its new location
                source = path.replace(before.get_path(), after.get_path())

        if ev.mask & pyinotify.IN_CREATE:
            self.bridge.on_create(path, source)
        elif ev.mask & pyinotify.IN_ATTRIB:
            self.bridge.on_attrib(path, source)
        elif ev.mask & pyinotify.IN_DELETE:
            self.bridge.on_delete(path)
        elif ev.mask & pyinotify.IN_MODIFY:
            if not ev.is_dir:
                # We wait for the final close after many changes
                # because we can't tell what part of the file was
                # changed and we'll do a full copy
                self.modified_files[path] = True
            else:
                # should not happen
                raise ResyncNeeded()
        elif ev.mask & pyinotify.IN_CLOSE_WRITE:
            # We were waiting for the close after many changes
            # Just check we saw change on the file
            if self.modified_files.get(path):
                del self.modified_files[path]
                self.bridge.on_change(path, source)
            else:
                log.info('Ignoring no-change wclose on %s' % path)
        elif ev.mask & pyinotify.IN_MOVED_FROM:
            self.rename_cookies[ev.cookie] = path
        elif ev.mask & pyinotify.IN_MOVED_TO:
            old = self.rename_cookies.get(ev.cookie)
            if not old:
                log.err('Rename with no matching cookie %s' % ev.cookie)
                raise ResyncNeeded()
            self.bridge.on_rename(old, path)
            del self.rename_cookies[ev.cookie]
            if self.modified_files.get(old):
                del self.modified_files[old]
                self.modified_files[path] = True
        else:
            log.err('unknown event : %s' % ev)

    def run(self):
        while self.running:
            try:
                self._run()
            except Exception, e:
                log.err('%s %s:%s' % (self, e.__class__, e,))

class ComparableEvent(pyinotify.Event):
    """This Event is comparable to another one using __eq__"""

    def __init__(self, ev):
        self.wd = ev.wd
        self.mask = ev.mask
        self.event_name = ev.maskname
        self.path = getattr(ev, 'path', None)
        self.name = getattr(ev, 'name', None)
        self.is_dir = getattr(ev, 'dir', None)
        self.cookie = getattr(ev, 'cookie', None)

    def get_path(self):
        return '%s/%s' % (self.path, self.name)

    def __eq__(self, other):
        return (
            getattr(other, 'path', None) == getattr(self, 'path', None) and
            getattr(other, 'name', None) == getattr(self, 'name', None) and
            getattr(other, 'mask', None) == getattr(self, 'mask', None) and
            getattr(other, 'cookie', None) == getattr(self, 'cookie', None)
        )

class GsyncProcessEvent(pyinotify.ProcessEvent):
    """This receives notifications from inotify and
    uses our protocol with repository objects (snapshot
    directories, local mirroring, etc.)"""

    def __init__(self):
        self.workers = []

    def add_worker(self, worker):
        log.debug('Adding worker %s' % worker)
        self.workers.append(worker)

    def process_default(self, ev):
        if ev.mask & pyinotify.IN_Q_OVERFLOW:
            log.warn('Queue overflow!')
            return

        for wk in self.workers:
            try:
                cev = ComparableEvent(ev)
                if wk.wants(cev):
                    log.debug('%s wants %s' % (wk, cev,))
                    try:
                        wk.queue.put(cev)
                    except QueueFull:
                        pass
            except Exception, e:
                log.err('%s %s:%s' % (self, e.__class__, e,))

class GsyncEngine:
    """This masks the inotify stuff and setups a GsyncProcessEvent
    instance."""

    def __init__(self):
        syslog.openlog('gsync')
        self.watch_manager = pyinotify.WatchManager()
        self.event_manager = GsyncProcessEvent()
        self.notify_thread = pyinotify.Notifier(self.watch_manager, self.event_manager)
        self.worker_threads = []
        self.snap_interval = 0
        self.snapshot_timer = None
        self.webserver_thread = None
        self.webserver = None
        self.running = None

    def snapshot(self):
        log.debug('snapshot timer triggered')
        self.snapshot_timer = threading.Timer(self.snap_interval, self.snapshot)
        try:
            for wk in self.worker_threads:
                wk.snapshot()
        except:
            log.error('Snapshot caught %s:%s' % (e.__class__, e,))
        self.snapshot_timer.start()
        log.debug('snapshot timer exiting')
        sys.exit()

    def setup_repository(self, name, conf):
        repo = Repository(name)

        try:
            repo.setup(conf)
        except Exception, e:
            log.err('Config failed for %s: %s' % (repo, e,))
            return None

        wt = GsyncWorker()
        wt.set_repository(repo)

        self.worker_threads.append(wt)

        if conf.get('export') and self.conf.get('webserver'):
            self.webserver.add_repository(repo, conf['export'])

        if repo.min_interval:
            if not self.snap_interval:
                self.snap_interval = repo.min_interval
            else:
                self.snap_interval = min(repo.min_interval, self.snap_interval)
                
        return wt 


    def stop(self):
        """Releases resources too"""
        log.info('Stopping engine')
        self.notify_thread.stop()

        for wt in self.worker_threads:
            wt.stop()

        if self.webserver_thread:
            self.webserver_thread.stop()

        if self.snapshot_timer:
            self.snapshot_timer.cancel()

        self.running = False

    def setup(self, conf):
        self.conf = conf
        log.info('Setting up')

        if self.conf.get('webserver'):
            self.webserver = GsyncWebserver()

        log.set_level(self.conf.get('loglevel', 'info'))

        for name, conf in self.conf['repositories'].iteritems():
            wt = self.setup_repository(name, self.conf['repositories'][name])
            if not wt:
                return False

            self.event_manager.add_worker(wt)
            for path in conf.get('monitor', []):
                log.debug('watching %s' % path)
                self.watch_manager.add_watch(
                    path, GSYNC_MASK, proc_fun=self.event_manager,
                    rec=True, auto_add=True
                )

        return True

    def run(self):
        if self.webserver:
            self.webserver.setup()
            self.webserver_thread = httpserver.runsimple(
                self.webserver.app.wsgifunc(), self.conf['webserver']
            )
            self.webserver_thread.start()
            log.debug('started %s' % self.webserver_thread)

        for wt in self.worker_threads:
            wt.start()
            log.debug('started %s' % wt)

        if self.snap_interval:
            self.snapshot_timer = threading.Timer(
                self.snap_interval, self.snapshot
            )
            self.snapshot_timer.start()

        log.info('Gandi sync daemon started')

        self.running = True
        
        while self.running:
            try:
                self.notify_thread.process_events()
                if self.notify_thread.check_events(60000):
                    self.notify_thread.read_events()
            except Exception, e:
                log.err('%s %s:%s' % (self, e.__class__, e,))

if __name__ == '__main__':
    conf = yaml.load(file('/etc/gandi/gsync.conf'))
    engine = GsyncEngine()
    try:
        if engine.setup(conf):
            engine.run()
    except (KeyboardInterrupt, Exception), e:
        engine.stop()
        log.info('exiting: %s:%s' % (e.__class__, e))
