"""Serves a repository. Implements the client to it"""

import posix
import posixpath
import remote
import time
from utils.filter import Filter, FilterChain
from utils.logger import log

import simulator
import versioning

class Repository:
    """A client has access to local files. The client stats
    file creations (symlinks, directories, files) and converts
    this following the repository protocol. A server only exports
    this protocol (allowing put/delete on files, attribute changes,
    snapshot handling, ..)"""
    
    repo_types = {
        'versioning': versioning.Repository,
        'remote': remote.Repository,
        'simulator': simulator.Repository
    }

    def __init__(self, name):
        self.name = name
        self.min_interval = 0

    def __repr__(self):
        return 'Repository(%s)' % self.name

    def setup(self, conf):
        log.info("Setting up %s" % (self.name,))

        self.conf = conf

        try:
            type = conf['type']
            path = conf['path']
        except KeyError, e:
            raise Exception('Missing %s from repository config' % e)

        try:
            self.repo = self.repo_types[type](path)
        except KeyError, e:
            raise Exception('Repository type %s does not exist' % e)

        ## Setup filters. If unspecified, defaults to True.
        ## If specified, defaults to False, unless default
        ## is also given.
        if not conf.get('filters'):
            self.filter = FilterChain(True)
        else:
            default = conf['filters'].get('default', False)
            self.filter = FilterChain(default)

        for re, action in conf.get('filters', {}).iteritems():
            self.filter.add(re, action)
        
        self.intervals = conf.get('intervals')
        if self.intervals:
            try:
                self.min_interval = min(
                    [ int(conf['delay']) for name, conf in 
                        self.intervals.iteritems() ]
                )
                log.debug('min interval for %s is %s' % (self, self.min_interval,))
            except KeyError, k:
                raise Exception('%s lacks interval delay' % self)
        else:
            self.min_interval = 0

    def alive(self):
        return self.repo.has('/')

    def wants(self, path):
        p = posixpath.normpath(path)

        # We want monitored paths that pass the filter rules,
        # If no monitor is defined, accept the file (exported setups)
        for m in self.conf.get('monitor', ['']):
            if p.startswith(posixpath.normpath(m)):
                return self.filter.wants(path)

        return False

    ## our file protocol:
    def create(self):
        """This creates the repository, not a file"""
        self.repo.create()
        # First snapshot increases version
        self.repo.snapshot()

    def has(self, path):
        return self.repo.has(path)

    def rename(self, a, b):
        """Rename a file or directory"""
        self.repo.rename(a, b)

    def add_from_fd(self, path, fd, metas=None, length=None):
        """Add or change the file"""
        self.repo.add_from_fd(path, fd, metas, length)

    def mkdir(self, path, metas=None):
        """Create a directory"""
        self.repo.mkdir(path, metas)

    def symlink(self, path, link, metas=None):
        self.repo.symlink(path, link, metas)
    
    def readlink(self, path, ver=None):
        self.repo.readlink(path, ver) 

    def update_metas(self, path, metas=None):
        self.repo.update_metas(path, metas)

    def remove(self, path):
        """Remove the file"""
        self.repo.remove(path)

    ## versionning stuff
    def _manage_intervals(self):
        """Some repositories handle snapshots, some do not. When they do,
        user can specify user-friendly intervals of snapshots to keep."""

        # Don't bother to manage anything if we only have one version
        if len(self.repo.versions()) <= 1:
            return

        now = self.repo.stat('/').st_mtime
        current = self.repo.current_version
        age_dict = {}
        for v in self.repo.versions():
            age_dict[v] = now - self.repo.stat('/', v).st_mtime

        self.repo.version_alias('current', current)

        aliases = self.repo.aliases()
        for name, conf in self.intervals.iteritems():
            delay = conf.get('delay')
            count = conf.get('count', 1)

            # Cleanup .soon entries
            soon = '.%s.soon' % (name,)
            if aliases.get(soon):
                soon_age = now - self.repo.stat('/', aliases[soon]).st_mtime
                if soon_age > (1.5 * delay):
                    log.info('Removing old candidate %s' % soon)
                    self.repo.remove_alias(soon)
                    aliases = self.repo.aliases()

            min_err = {}
            this_period = {}
            for v, age in age_dict.iteritems():
                n = 1.00 * age / delay
                if not int(n):
                    continue

                log.debug('estimating %s:%s with %s' % (name, n, v,))

                err = 1.00 * abs(age - (delay * n)) / (delay * n)
                log.debug('candidate %s has err %s' % (v, err,))
                if int(n) <= count and err < min_err.get(int(n), 1):
                    log.debug(' -> %s is best' % (v,))
                    min_err[int(n)] = err
                    this_period[int(n)] = v
            
            for n, v in this_period.iteritems():
                if aliases.get('%s.%s' % (name, n,)) != v:
                    log.debug('best candidate %s => %s' % (n, v,))
                    self.repo.version_alias('%s.%s' % (name, n,), v)
                    self.repo.version_alias(soon, current)
            else:
                if soon not in aliases:
                    log.info('Setting up candidate for %s' % name)
                    self.repo.version_alias(soon, current)

        aliases = self.repo.aliases()
        aliased_versions = [ aliases[x] for x in aliases ]
        for v in self.repo.versions():
            if v not in aliased_versions:
                self.repo.remove_version(v)

    def snapshot(self):
        self.repo.snapshot()
        if self.intervals:
            self._manage_intervals()

    def aliases(self):
        return self.repo.aliases()

    def versions(self):
        return self.repo.versions()

    def use(self, v):
        return self.repo.use(v)

    def stat(self, path, ver=None):
        return self.repo.stat(path, ver)

    def get_fd(self, path, ver=None):
        return self.repo.get_fd(path, ver)

    def list(self, path, ver=None):
        return self.repo.list(path, ver)

    def last_version(self):
        last_version = max(self.repo.versions(), key=lambda x: self.repo.stat('/', x).st_mtime)
        return last_version
