import errno
import httplib
import urllib
import base64
import re
import string
import socket
from urlparse import urlparse
from time import gmtime, strftime, sleep
from web.utils import Storage

from utils.logger import log

class HTTPRemoteConnectionError(Exception):
    """
    Class for handling errors from http connection
    """
    def __init__(self, code='', message='', resource=''):
        Exception.__init__(self)
        self.code = code
        self.message = message
        self.resource = resource

    def __str__(self):
        return 'Error: %s - %s\n%s' % (self.code, self.message, self.resource)
    
    def __repr__(self):
        return '%s: %s\n%s' % (self.code, self.message, self.resource)

class HTTPRemoteConnection(object):
    """
    Class for handling connection to a http server
    """
    def __init__(self, url):
        self.url = url
        o = urlparse(url)
        self.login = o.username
        self.password = o.password
        if o.scheme == 'https':                                     
            self._conn = httplib.HTTPSConnection(o.hostname, o.port)
        else:                                                      
            self._conn = httplib.HTTPConnection(o.hostname, o.port)

        self.path = o.path.lstrip('/').rstrip('/')
           
    def clone(self):
        """
        returns a new connection to the http server
        """
        return HTTPRemoteConnection(self.url) 

    def __getattr__(self, attr): 
        method = attr.upper()
        if method not in ['GET', 'PUT', 'HEAD', 'POST', 'DELETE']:
            raise Exception('Unsupported method')
        def f(path='', send_io=None, params=None, headers=None):
            path = '/repositories/%s/%s' % (self.path, path)
            r = re.compile('/+')
            path = r.sub('/', path)
            length = None
            if isinstance(headers, dict) and headers.has_key("Content-Length"):
                length = headers["Content-Length"]
            elif send_io is not None:
                length = self._io_len(send_io)
            headers = self._headers(method, headers=None, length=length)
            def do_connect():
                self._conn.putrequest(method, path + self._params(params))
                for k, v in headers.items():
                    self._conn.putheader(k, v)
                self._conn.endheaders()
            retry = False
            try:
                do_connect()
            except socket.error, e:
                #Try again on Broken Pipe
                if e[0] == 32:
                    retry = True
            if retry:
                do_connect()
            if send_io is not None:
                data = send_io.read(httplib.MAXAMOUNT)
                while len(data) > 0:
                    self._conn.send(data)
                    data = send_io.read(httplib.MAXAMOUNT)
                send_io.read()
            try:
                r = self._conn.getresponse()
            except httplib.ResponseNotReady, e:
                log.err("[HTTPConnectionError] %s:%s" % (e.__class__, e))
            except Exception, e:
                log.err("[HTTPConnectionError] %s:%s" % (e.__class__, e))
            else:
                log.debug('Status %s' % r.status)
                if r.status < 200 or r.status > 299:
                    raise HTTPRemoteConnectionError(
                        code=r.status, message=r.reason, 
                        resource=r.read()
                    )
                if method == 'HEAD':
                    return r.getheaders()
                if method != 'GET':
                    r.read()
                return r
        return f
    
    def _headers(self, method, length=None, headers=None):
        if not headers:
            headers = {}
        if length is not None:
            headers["Content-Length"] = length

        if self.login and self.password:
            headers['Authorization'] = 'Basic ' +\
                string.strip(base64.encodestring(
                    '%s:%s' % (self.login,self.password))
                )
        headers["Date"] = strftime("%a, %d %b %Y %H:%M:%S GMT", gmtime())
        headers['User-Agent'] = 'Gsync User Agent'
        return headers

    def _params(self, params):
        p = ''
        if params: 
            p = '?' + urllib.urlencode(params)
        return p

    def _io_len(self, io):
        if hasattr(io, "len"):
            return io.len
        o_pos = io.tell()
        io.seek(0, 2)
        length = io.tell() - o_pos
        io.seek(o_pos, 0)
        return length

class Repository:
    def __init__(self, path):
        self.path = path
        self.current_version = None
        self.conn = HTTPRemoteConnection(path)
    
    def create(self):
        """
        """
        pass

    def mkdir(self, path, metas=None):
        """
        Directory creation request
        """
        if metas is None: metas = {}
        try:
            metas['-p'] = 1
            return self.conn.clone().put(path, params=metas)
        except HTTPRemoteConnectionError, e:
            if e.code == 404:
                raise EnvironmentError(errno.ENOENT, 'No such file')
        except Exception, e:
            log.err("[HTTPConnectionError] %s:%s" % (e.__class__, e))

    def add_from_fd(self, path, fd, metas=None, length=None):
        """
        File creation request
        """
        try:
            return self.conn.clone().put(path, send_io=fd, params=metas)
        except HTTPRemoteConnectionError, e:
            if e.code == 404:
                raise EnvironmentError(errno.ENOENT, 'No such file')
        except Exception, e:
            log.err("[HTTPConnectionError] %s:%s" % (e.__class__, e))

    def remove(self, path):
        """
        File or directory deletion request
        """
        try:
            return self.conn.clone().delete(path)
        except Exception, e:
            log.err("[HTTPConnectionError] %s:%s" % (e.__class__, e))

    def symlink(self, path, symlink, metas=None):
        """
        Symbolic link creation request
        """
        try:
            return self.conn.clone().post(path, params={'symlink': symlink})
        except Exception, e:
            log.err("[HTTPConnectionError] %s:%s" % (e.__class__, e))

    def update_metas(self, path, metas):
        """
        Stat file modification request
        """
        try:
            return self.conn.clone().post(path, params=metas)
        except HTTPRemoteConnectionError, e:
            if e.code == 404:
                raise EnvironmentError(errno.ENOENT, 'No such file')
        except Exception, e:
            log.err("[HTTPConnectionError] %s:%s" % (e.__class__, e))

    def rename(self, patha, pathb):
        """
        File or repository renaming request
        """
        try:
            return self.conn.clone().post(patha, params={'name': pathb})
        except Exception, e:
            log.err("[HTTPConnectionError] %s:%s" % (e.__class__, e))

    def list(self, path='/', ver=None):
        """
        Directory listing request
        """
        try:
            res = self.conn.clone().get(path, params={'version': ver})
            res = res.split('\n')
            return res
        except Exception, e:
            log.err("[HTTPConnectionError] %s:%s" % (e.__class__, e))

    def snapshot(self):
        """
        """
        pass

    def has(self, path):
        """
        checks file or directory existence
        """
        try:
            res = self.conn.clone().head(path)
            if res: return True
        except Exception, e:
            log.err("[HTTPConnectionError] %s:%s" % (e.__class__, e))
        return False

    def use(self, ver=None):
        self.current_version = ver

    def stat(self, name, ver=None):
        attrs = ['mode', 'ino', 'dev', 'nlink', 'uid', 
                'gid', 'size', 'atime', 'mtime', 'ctime']
        try:
            headers = self.conn.clone().head(name, params={'version': ver})
        except Exception, e:
            log.err("[HTTPConnectionError] %s:%s" % (e.__class__, e))
        if headers:
            headers = dict(headers)
            stat = dict([ ('st_' + i,int(headers.get('x-gsync-stat-' + i, 0))) for i in attrs]) 
            stat = Storage(**stat)
            return stat

    def aliases(self):
        aliases = {}
        try:
            res = dict(self.conn.clone().head('/?aliases'))
            if res:
                for k,v in res.items():
                    regex = re.compile('^' + 'x-gsync-aliases' + '(-(.*))?$')
                    m = regex.match(k)
                    if m and m.group(2):
                        aliases[m.group(2)] = int(v)
        except Exception, e:
            log.err("[HTTPConnectionError] %s:%s" % (e.__class__, e))
        return aliases

    def versions(self):
        versions = []
        try:
            res = dict(self.conn.clone().head('/?versions')).get('x-gsync-versions', '')
            versions = map( lambda x : int(x), res.split(','))
        except Exception, e:
            log.err("[HTTPConnectionError] %s:%s" % (e.__class__, e))
        return versions
