"""
Provides an interface to directory tree versionning trough
unix hardlinks.
"""

import datetime
import errno
import os
import posix
import posixpath
import stat

from utils.logger import log

class Repository:
    def __init__(self, path):
        self.path = path
        self.current_version = None

    def create(self):
        try:
            posix.umask(022)
            posix.mkdir(self.path)
        except EnvironmentError, e:
            if e.errno == errno.EEXIST:
                ## Just cleanup aliases
                aliases = self.aliases()
                return
            raise

    def aliases(self):
        """List the current aliases and cleanup broken ones"""
        li = {}
        for str in os.listdir(self.path):
            try:
                path = '%s/%s' % (self.path, str,)
                st = posix.lstat(path)
                if stat.S_ISLNK(st.st_mode):
                    try:
                        v = posix.readlink(path)
                        # Check this version exists
                        if not posix.access(v, 0):
                            posix.unlink(path)
                        elif v.startswith(self.path):
                            li[str] = int(v.strip('/').split('/')[-1])
                    except EnvironmentError, e:
                        if e.errno == errno.ENOENT:
                            posix.unlink(path)
            except:
                pass
        return li
                
    def versions(self):
        li = []
        try:
            for ver in os.listdir(self.path):
                try:
                    ver = int(ver)
                    if self._valid(ver):
                        li.append(ver)
                    else:
                        self.remove_version(ver)
                except:
                    pass
        except EnvironmentError, e:
            if e.errno == errno.ENOENT: pass
        return li
    
    def use(self, ver):
        if not self._valid(ver):
            raise Exception('Version %d is broken' % ver)
        self.current_version = ver

    def _make_version(self, ver=None):
        if ver is None:
            ver = self.current_version
        posix.mkdir(self._versionned('/', ver))
        self.current_version = ver

    def _lock(self, ver=None):
        if ver is None:
            ver = self.current_version
        return '%s/.%03d.invalid' % (self.path, ver,)

    def _valid(self, ver=None):
        if not stat.S_ISDIR(posix.stat(self._versionned('/', ver)).st_mode):
            return False
        return not posixpath.exists(self._lock(ver))

    def _begin(self, ver=None):
        """Mark invalid before a non atomic operation"""
        try:
            posix.mkdir(self._lock(ver))
        except:
            pass

    def _commit(self, ver=None):
        """Remove invalid after a non atomic operation"""
        try:
            posix.rmdir(self._lock(ver))
        except:
            pass
        
    def _sanitize(self, name):
        n = posixpath.normpath(name)
        return n

    def _versionned(self, name, ver=None):
        if ver is None:
            ver = self.current_version

        name = self._sanitize(name)

        return '%s/%03d/%s' % (
            self.path,
            ver,
            name.lstrip('/')
        )

    def _unversionned(self, name):
        if name.startswith(self.path):
            name = name.replace(self.path, '', 1)
            (ver, path) = name.lstrip('/').split('/', 1)
            try:
                v = int(ver)
                return '/%s' % path
            except ValueError, e:
                pass

        raise Exception('Bad file name %s' % name)

    def list(self, path='/', ver=None):
        if not ver: ver = self.current_version
        for f in os.listdir(self._versionned(path, ver)):
            yield f

    def walk(self, path='/', ver=None, t=True):
        for (dirname, dirs, files) in os.walk(self._versionned(path, ver), t):
            dirname = self._unversionned(dirname)
            yield dirname, dirs, files

    def _hardlinks(self, v1):
        for (dirname, dirs, files) in self.walk('/', v1):
            for dir in dirs:
                path = '%s/%s' % (dirname, dir,)
                st = posix.stat(self._versionned(path, v1))
                self.mkdir(path, {
                        'uid': st.st_uid,
                        'gid': st.st_gid,
                        'mtime': st.st_mtime,
                        'atime': st.st_atime,
                        'mode': st.st_mode
                    }
                )
            for file in files:
                path = '%s/%s' % (dirname, file,)
                posix.link(
                    self._versionned(path, v1),
                    self._versionned(path)
                )

    def remove_version(self, v):
        self._begin(v)
        for (dirname, dirs, files) in self.walk('/', v, False):
            for f in files + dirs:
                self.remove('%s/%s' % (dirname, f), v)
            self.remove(dirname, v)
        self._commit(v)

    def remove_alias(self, name):
        os.unlink('%s/%s' % (self.path, name))

    def version_path(self, v):
        return self._versionned('/', v)

    def version_alias(self, name, ver=None):
        sym_path = '%s/%s' % (self.path, name)

        try:
            posix.unlink(sym_path)
        except:
            pass

        posix.symlink(self._versionned('/', ver), sym_path)

    def remove(self, name, ver=None):
        fname = self._versionned(name, ver)
        try:
            st = posix.lstat(fname)
        except OSError, e:
            if e.errno == errno.ENOENT:
                return

        if stat.S_ISDIR(st.st_mode):
            posix.rmdir(fname)
        else:
            posix.unlink(fname)

    def snapshot(self):
        """Close the current version (or do nothing)
        Create a new one with hardlinks (or not)
        Increments (or not) current_version """
        old = self.current_version
        if old is None:
            self.current_version = 0
        self.current_version += 1
        self._begin()
        self._make_version()
        if old is not None:
            self._hardlinks(old)
        self._commit()

    def update_metas(self, name, metas, ver=None):
        fname = self._versionned(name, ver)
        lnk = stat.S_ISLNK(self.stat(name).st_mode)
        try:
            if not lnk and (metas.get('mode')):
                posix.chmod(fname, stat.S_IMODE(metas.get('mode')))

            if not lnk and (metas.get('atime') or metas.get('mtime')):
                posix.utime(fname, (metas.get('atime'), metas.get('mtime')))

            if metas.get('uid') or metas.get('gid'):
                posix.lchown(fname, metas.get('uid', -1), metas.get('gid', -1))
        except Exception, e:
            log.err('Got %s in update_metas' % e)
            pass
       
    def stat(self, name, ver=None):
        return posix.lstat(self._versionned(name, ver))
    
    def readlink(self, name, ver=None):
        return self._unversionned(
            posix.readlink(self._versionned(name, ver))
        )

    def mtime(self, name, ver=None):
        return datetime.datetime.fromtimestamp(self.stat(name, ver).st_mtime)

    def _mkdirs(self, name, metas=None):
        """name must be a directory!
        Prepare parent directories for the first import of a deep path"""
        src = posixpath.normpath(name)
        for i in range(src.count('/'), 0, -1):
            p = src.rsplit('/', i)[0]
            if not self.has(p):
                self.mkdir(p, metas)

    def add_from_fd(self, name, fd, metas=None, length=None):
        """Adds (or update) the given file to the current version."""
        fname = self._versionned(name)

        self._begin()

        try:
            posix.unlink(fname)
        except:
            pass

        f = file(fname, 'w')

        chunksize = 4096
        # Actual copy
        #We return if we know the file length
        if length is not None:
            while 1:
                remain = length - f.tell()
                if remain <= 0: break
                chunk = fd.read(min(chunksize, remain))
                if not chunk: break
                f.write(chunk)
            f.close()
            if metas:
                self.update_metas(name, metas)

            self._commit()
            return
        else:
            buf = fd.read(chunksize)
            while buf:
                f.write(buf)
                buf = fd.read(chunksize)

            f.close()

            if metas:
                self.update_metas(name, metas)

            self._commit()

    def rename(self, a, b, ver=None):
        """Rename file a to b"""
        va = self._versionned(a)
        vb = self._versionned(b)
        return posix.rename(va, vb)
        
    def symlink(self, orig, dest, metas=None):
        """Beware that symlinks will point within the version"""
        v_orig = self._versionned(orig)

        if dest.startswith('/'):
            v_dest = self._versionned(dest)
        else:
            v_dest = dest

        self._begin()
        posix.symlink(v_dest, v_orig)
        if metas:
            self.update_metas(orig, metas)
        self._commit()

    def mkdir(self, name, metas=None):
        """Special meta "-p" means recursive dir creation"""
        fname = self._versionned(name)
        self._begin()
        try:
            posix.mkdir(fname)
        except OSError, e:
            if e.errno == errno.ENOENT and metas.get('-p'):
                # build directories
                self._mkdirs(name, metas)
                posix.mkdir(fname)

        if metas:
            self.update_metas(name, metas)
        self._commit()

    def has(self, name):
        try:
            if self.stat(name):
                return True
        except:
            pass
        return False

    def get_fd(self, name, ver=None):
        fname = self._versionned(name, ver)
        return file(fname, 'r')

# vim: et:ts=4
