"""
Allows for name filter definition, and the definition of a chain
of filters.
"""

import re
import fnmatch
import unittest

class Filter:
    def __init__(self, regexp, keep):
        self.re_text = regexp
        try:
            self.regexp = re.compile(regexp)
            self.glob = None
        except re.error, e:
            self.regexp = None
            self.glob = regexp
        self.keep = keep
    
    def wants(self, name):
        if self.regexp:
            if self.regexp.search(name):
                return self.keep
        elif self.glob:
            if fnmatch.fnmatch(name, self.glob):
                return self.keep
        
    def __repr__(self):
        return 'Filter: %s -> %s' % (self.re_text, self.keep)

class FilterChain:
    def __init__(self, default=False):
        self.filters = []
        self.default = default

    def add(self, regexp, keep):
        self.filters.append(Filter(regexp, keep))

    def wants(self, name):
        for f in self.filters:
            ret = f.wants(name)
            if ret is not None:
                return ret

        return self.default

class FilterTestCase(unittest.TestCase):
    def setUp(self):
        self.complex_chain_no = FilterChain(False)
        self.complex_chain_yes = FilterChain(True)
        for c in self.complex_chain_yes, self.complex_chain_no:
            c.add('.*\.log$', False)
            c.add('/mysql/', False)
            c.add('^/this_dir_ok/', True)
            c.add('^/this_dir_no/', False)

    def test_invalid_regexp(self):
        try:
            f = Filter('^(', False)
            self.fail('Invalid regexp passed')
        except:
            pass

    def test_basic_regexp(self):
        f = Filter('^works$', True)
        self.assert_(f.wants('works'))
        self.assert_(not f.wants('work'))

    def test_default_for_chain(self):
        self.assert_(self.complex_chain_yes.wants('/anything'))
        self.assert_(not self.complex_chain_no.wants('/anything'))

    def test_glob(self):
        f = Filter('*.sql', False)
        self.assert_(not f.wants('/var/tmp/lib.sql'))
        f = Filter('*.sql', True)
        self.assert_(f.wants('/var/tmp/lib.sql'))
        self.assert_(not f.wants('/var/tmp/lib.tmp'))
        self.assert_(not f.wants('/libsql/lib.tmp'))
        self.assert_(not f.wants('/lib.sql/lib.tmp'))

    def test_chain(self):
        f = FilterChain(False)
        self.assert_(self.complex_chain_no.wants('/this_dir_ok/test/file'))
        self.assert_(not self.complex_chain_no.wants('/this_dir_ok/a.log'))
        self.assert_(self.complex_chain_no.wants('/this_dir_ok/a.log/'))
        self.assert_(not self.complex_chain_no.wants('/this_dir_ok/mysql/test.db'))
        self.assert_(not self.complex_chain_yes.wants('/this_dir_ok/mysql/test.db'))
        self.assert_(not self.complex_chain_no.wants('/this_dir_no/'))

if __name__ == '__main__':
    unittest.main()
